package wing.fx.dialog;

import javafx.collections.FXCollections;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonBase;
import javafx.scene.control.Label;
import javafx.scene.layout.Pane;
import javafx.stage.Stage;
import javafx.stage.Window;
import org.tbee.javafx.scene.layout.MigPane;

import java.util.*;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * @author Mikael Grev, MiG InfoCom AB
 *         Date: 14-06-22
 *         Time: 13:10
 */
public class Dialog<R>
{
	private static final ButtonType[] DEF_BUTTONS = new ButtonType[] {ButtonType.OK, ButtonType.CANCEL};

	private final MigPane pane = new MigPane();

	private final ArrayList<Node> panes = new ArrayList<>();
	private final ArrayList<String> constraints = new ArrayList<>();

	private final DialogContent<R> content;
	private Stage dialogStage;

	private ButtonType[] buttons = null;
	private IdentityHashMap<Node, ButtonType> buttonMap = null;
	private Consumer<R> applyConsumer;

	private Node pressedButton;
	private ButtonType pressedButtonType;
	private Optional<R> result = Optional.empty();

	private Dialog(DialogContent<R> content)
	{
		this.content = content;
	}

	public Node getPressedButton()
	{
		return pressedButton;
	}

	public ButtonType getPressedButtonType()
	{
		return pressedButtonType;
	}

	public Optional<R> getResult()
	{
		return result;
	}

	public static <E> Dialog<E> create(DialogContent<E> content)
	{
		return new Dialog<>(content);
	}

	public static Dialog<ButtonType> createMessage(String text)
	{
		return new Dialog<>(new MessageContent(text));
	}

	public static Dialog<String> createSingleLineInput(String label)
	{
		return new Dialog<>(new SingleLineInput(label, null, null, -1, null));
	}

	public static Dialog<String> createSingleLineInput(String label, Predicate<String> validator)
	{
		return new Dialog<>(new SingleLineInput(label, null, null, -1, validator));
	}

	public static <V> Dialog<V> createSingleChoice(String label, Collection<V> choices)
	{
		return createSingleChoice(label, choices, null);
	}

	public static <V> Dialog<V> createSingleChoice(String label, Collection<V> choices, Function<V, String> toString)
	{
		return new Dialog<>(new SingleChoice<>(label, FXCollections.observableArrayList(choices), Optional.empty(), toString));
	}

	public Dialog<R> applyTo(Consumer<R> consumer)
	{
		this.applyConsumer = consumer;
		return this;
	}

	public Dialog<R> masthead(String text)
	{
		return masthead(text, null);
	}

	public Dialog<R> masthead(String text, Node gfx)
	{
		MigPane masthead = new MigPane();
		masthead.add(new Label(text));
		if (gfx != null)
			masthead.add(gfx);

		panes.add(0, masthead); // Always first
		constraints.add(0, "dock north");

		return this;
	}

	public Dialog<R> buttons(ButtonType... buttons)
	{
		this.buttons = buttons;
		return this;
	}

	public Dialog<R> addButton(ButtonBase button, ButtonType type)
	{
		if (buttonMap == null)
			buttonMap = new IdentityHashMap<>();

		buttonMap.put(button, type);
		button.setOnAction(event -> buttonPressed((Node) event.getSource(), type));

		return this;
	}

	public Dialog<R> addPane(Node pane, String migPaneCC)
	{
		panes.add(pane);
		constraints.add(migPaneCC);

		return this;
	}

	public Optional<R> show()
	{
		return show(null);
	}

	public Optional<R> show(Window owner)
	{
		pane.add(createButtonRow(), "dock south");
		for (int i = 0; i < panes.size(); i++)
			pane.add(panes.get(i), constraints.get(i));

		pane.add(content.getContentNode(), "dock center");

		dialogStage = new Stage();
		if (owner != null)
			dialogStage.initOwner(owner);
		Scene scene = new Scene(pane);
		dialogStage.setScene(scene);

		content.setDialog(this);

		dialogStage.showAndWait();

		return result;
	}

	public Optional<ButtonBase> getButton(ButtonType type)
	{
		Node node = pane.lookup("#" + type);
		return node instanceof ButtonBase ? Optional.of((ButtonBase) node) : Optional.empty();
	}

	public List<ButtonBase> getApplyButtons()
	{
		return Arrays.stream(ButtonType.values())
			.filter(ButtonType::isApply)
			.map(type -> {
				Node node = pane.lookup("#" + type);
				return node instanceof ButtonBase ? (ButtonBase) node : null;
			})
			.filter(button -> button != null)
			.collect(Collectors.toList());
	}


	private Pane createButtonRow()
	{
		MigPane buttonRow = new MigPane("nogrid");

		if (buttonMap != null)
			buttonMap.entrySet().stream().forEach(entry -> buttonRow.add(entry.getKey(), getCcFromType(entry.getValue())));

		if (buttons != null)
			Arrays.stream(buttons).forEach(type -> buttonRow.add(createDefaultButton(type), getCcFromType(type)));

		if (buttonRow.getChildren().isEmpty())
			Arrays.stream(DEF_BUTTONS).forEach(type -> buttonRow.add(createDefaultButton(type), getCcFromType(type)));

		return buttonRow;
	}

	private Node createDefaultButton(ButtonType type)
	{
		Button button = new Button(getNameFromType(type));
		button.setOnAction(event -> buttonPressed((Node) event.getSource(), type));
		button.setId(type.toString());
		return button;
	}

	private void buttonPressed(Node button, ButtonType type)
	{
		pressedButton = button;
		pressedButtonType = type;

		Optional<R> resO = null;

		if (content.shouldApply(button, type)) {
			resO = content.getResult(type);
			if (applyConsumer != null)
				resO.ifPresent(applyConsumer::accept);
		}

		if (content.shouldClose(button, type)) {
			result = resO != null ? resO : content.getResult(type);
			dialogStage.hide();
		}
	}

	private String getNameFromType(ButtonType type)
	{
		switch (type) {
			case APPLY: return "Apply";
			case OK: return "OK";
			case CANCEL: return "Cancel";
			case YES: return "Yes";
			case NO: return "No";
			case FINISH: return "Finish";
			case PREVIOUS: return "< Previous";
			case NEXT: return "Next >";

			case CUSTOM:
			default: return "???";
		}
	}

	private static String getCcFromType(ButtonType value)
	{
		switch (value) {
			case APPLY: return "tag apply, wmin button";
			case OK: return "tag ok, wmin button";
			case CANCEL: return "tag cancel, wmin button";
			case YES: return "tag yes, wmin button";
			case NO: return "tag no, wmin button";
			case FINISH: return "tag finish, wmin button";
			case PREVIOUS: return "tag back, wmin button";
			case NEXT: return "tag next, wmin button";

			case CUSTOM:
			default: return "wmin button";
		}
	}
}